package FluxY_Lib::MS::DataClass_AA;

################################################################################
# This class provides support for amino acid compound fragments, derivatized
# using TBDMS.
#
# This class inherits basic functionality from the DataClass parent class,
# initializing it with the Amino Acid compounds, the atom list and compound
# name. It extends the basic DataClass to add support for derivatized fragments.
# As such it over-rides the GetCarbonInfo() and GetAtomCount() methods, as well
# as implementing the required PrintMassTables(). Seven additional helper
# classes, GenerateAtomList(), GenerateMassList(), GenerateConflictList(), 
# GetTotalAtomList(), GetMassList(), GetConflictList() and GetCarbonBackbone(),
# are also implemented.
#
################################################################################

use strict;
use warnings;

use FluxY_Lib::Chemistry::atomInfo; # User Lib - Atomic Masses.
use FluxY_Lib::Math::Math_Utils;    # User Lib - Simple math functions.

use FluxY_Lib::MS::DataClass;       # User Lib - The common parent class.
use FluxY_Lib::MS::DAgent;          # User Lib - Provides Derivatizing Agent Info.

our @ISA = qw(FluxY_Lib::MS::DataClass);    # Inherits from DataClass


# List of the atoms found in amino acids and subsequently considerred for
# natural abundance. 
# Note: Si is included for convenience (commonly found in derivatizing agents).
my @atomList = ('C','H','N','O','S','Si');
my @atomMass = ();

# Each of the following atom counts is an array of the number of each corresponding
# atom relative to the order specified in the atomList, i.e. C, H, N, O, S and Si.

# Hash of compound data, (first two array elements are required).
# 1. The compound names (amino acids), in lower case (for standard comparison).
# 2. The atom count for each Amino Acid (side group, R, + backbone) compound.
# 3. The number of silyl groups for each Amino Acid.
# 4. Number of hydrogens added to the m302s side chain.
#    Multiplier - the number of dAgents that attach to the Amino Acid.
my %compounds = (
   ala => ["alanine",       [3,7,1,2,0,0],   2, 0],
   gly => ["glycine",       [2,5,1,2,0,0],   2, 0],
   val => ["valine",        [5,11,1,2,0,0],  2, 0],
   leu => ["leucine",       [6,13,1,2,0,0],  2, 0],
   ile => ["isoleucine",    [6,13,1,2,0,0],  2, 0],
   pro => ["proline",       [5,9,1,2,0,0],   2, 0],
   cys => ["cysteine",      [3,7,1,2,1,0],   3, 0],
   met => ["methionine",    [5,11,1,2,1,0],  2, 0],
   ser => ["serine",        [3,7,1,3,0,0],   3, 0],
   thr => ["threonine",     [4,9,1,3,0,0],   3, 0],
   phe => ["phenylalanine", [9,11,1,2,0,0],  2, 0],
   asp => ["aspartic acid", [4,7,1,4,0,0],   3, 0],
   glu => ["glutamic acid", [5,9,1,4,0,0],   3, 0],
   asn => ["asparagine",    [4,8,2,3,0,0],   3, 0],
   lys => ["lysine",        [6,14,2,2,0,0],  3, 0],
   gln => ["glutamine",     [5,10,2,3,0,0],  3, 0],
   arg => ["arginine",      [6,14,4,2,0,0],  5, 0],
   his3 => ["histidine",    [6,9,3,2,0,0],   3, 1],   # 3 refers to the number of silyl groups.
#   his2 => ["histidine",    [6,10,3,2,0,0],  2, 0],   # 2 refers to the number of silyl groups.
   tyr => ["tyrosine",      [9,11,1,3,0,0],  3, 0],
   trp => ["tryptophan",    [11,12,2,2,0,0], 3, 0],
   );

# A 2D Hash of compound masses, indexed by compunds key and frags keys.
# Each element is the truncated mass value for the given compound/fragment.
my %massList = ();
# A 2D Hash of fragment conflicts, indexed by compunds key and frags keys.
my %conflictList = ();
# A 2D Hash of fragment atom lists, indexed by compunds key and frags keys.
my %totalAtomList = ();
# The name and atom count for the derivatizing agent.
my $dAgentName = "";
my @dAgent     = ();

# Hash of fragment atom reductions for the major fragmented mass groups:
# - current implementation is for TBDMS only
my %frags = (
   m000  => [0,0,0,0,0,0],   # M - molecular mass, added for convenience.
   m015  => [1,3,0,0,0,0],   # M-15
   m057  => [4,9,0,0,0,0],   # M-57
   m085  => [5,9,0,1,0,0],   # M-85
   m159  => [7,15,0,2,0,1],  # M-159
   m302s => [14,32,1,2,0,2], # M-302 (loss of side chain)
   );

my @m302 = (14,32,1,2,0,2);  # M=302 is a common fragment to all AA (loss of R).


################################################################################
# Class constructor
#
# Used to instantiate an object for this class.
#
################################################################################
sub new {
   my ($class, $dAgent) = @_;

   # Initialize the atom mass vector.
   @atomMass = &GetElementMasses(@atomList);

   # If not specified, set dAgent to TBDMS by default, otherwise check that the
   # derivatizing agent is TBDMS, or terminate.
   if ( !(defined($dAgent)) ) {
      $dAgent = "TBDMS";
   } elsif ( lc($dAgent) ne 'tbdms' ) {
      print "\n\n\tThe specified derivatizing agent, '" . $dAgent . "', could not be used.\n\n";
      print "\tCurrently only TBDMS is supported for derivatized amino acids.\n\n";
   
      exit -1;
   }
   
	my ($dName, $atomCount, $atomList) = &GetDAgent($dAgent);
	
	$dAgentName     = $dName;
	my @atomCountDA = @{$atomCount};
	my @atomListDA  = @{$atomList};

   # Ensure that the derivatizing agent is in the same order as the atomList.
   foreach my $atom (@atomList) {
      my $found = 0;
      for (my $i=0; $i<@atomListDA; $i++) {
         if ( $atom eq $atomListDA[$i] ) {
            $found = 1;
            push(@dAgent, $atomCountDA[$i]);
            last;
         }
      }
      if ( !$found ) { push(@dAgent, 0); }
   }
   
   # Initialize the total atom list.
   GenerateAtomList();
   # Initialize the compound mass list.
   GenerateMassList();
   # Initialize the compound conflict list.
   GenerateConflictList();

   # Call the constructor of the parent class, Person.
   my $self = $class->SUPER::new(\%compounds, \@atomList, "Amino Acids (AA)");
   
   bless $self, $class;

   return $self;
}


########################################################
#                                                      #
#   Over-ridden DataClass methods.                     #
#                                                      #
#   - to support the derivating agent and fragments.   #
#                                                      #
########################################################

################################################################################
# Subroutine: GetCarbonInfo
#
# Returns the number of carbon atoms, and the relative positional carbon string
# for the given fragment.
#
# NOTE: When 2 or more fragments have the same mass they cannot be distinguished
#       by this subroutine. Data is returned for the first fragment occuring in
#       alphabetical order (based on the fragment identifier key sort).
#
################################################################################
sub GetCarbonInfo {
   my ($self, $compound, $fragMass) = @_;
   
	my ($firstC, $lastC) = GetCarbonBackbone($compound, $fragMass, $self->GetNumCarbons($compound));

	if ( !defined($firstC) ) { return; }

	my $fragAtomStr = "";
	my $numCarbons  = $lastC - $firstC + 1;
		
	while ( $firstC < $lastC ) {
		$fragAtomStr .= "$firstC,";
		$firstC++;
	}         
	$fragAtomStr .= "$lastC";
		
	return ($numCarbons, $fragAtomStr);
}


################################################################################
# Subroutine: GetAtomCount
#
# Returns the atom count corresponding to the specific Amino Acid compound, and 
# derivitized fragment mass.
#
# NOTE: When 2 or more fragments have the same mass they cannot be distinguished
#       by this subroutine. Data is returned for the first fragment occuring in
#       alphabetical order (based on the fragment identifier key sort).
#
################################################################################
sub GetAtomCount {
   my ($self, $compound, $fragMass) = @_;
   
   foreach my $frag (sort keys %{$totalAtomList{$compound}}) {      
      if ($fragMass == $massList{$compound}{$frag}) {         
         return @{$totalAtomList{$compound}{$frag}};
      }
   }
   
   return;
}


######################################
#                                    #
#   Implemented DataClass methods.   #
#                                    #
######################################

################################################################################
# Subroutine: PrintMassTables
#
# Prints a table containing the atom count / fragment mass for each Amino Acid.
# The table contains a row for the derivatized amino acid and each of the major 
# fraction groups.
#
################################################################################
sub PrintMassTables {
   my ($self) = @_;

   my $outFile = "AA_Data_Tables.txt";
	
   open (FILE, ">" . $outFile) or FileSystemError(900, $outFile, $!);
   
   print FILE "The derivatizing agent is: " . $dAgentName . "\n";
   
   foreach my $compound (sort keys %compounds) {
      print FILE "\nFor: " . $compound . "\n";
   
      print FILE "\t" . join("\t", @atomList) . "\tMass\tC-Backbone\n";
      
		my $firstC;
      my $lastC;
      foreach my $frag (sort keys %{$totalAtomList{$compound}}) {
         my @totAtoms = @{$totalAtomList{$compound}{$frag}};
			
         my $sFlag = 0;
			if ( $frag eq "m302s") { $sFlag = ${$compounds{$compound}}[3]; }
         
			if ( $frag eq "m302") {
			   ($firstC, $lastC) = (1,2);
			} else {
		      ($firstC, $lastC) = GetCarbonBackbone($compound, $massList{$compound}{$frag}, $self->GetNumCarbons($compound));
			}
			
         print FILE $frag;
         if ( $sFlag ) { print FILE "**"; }
         print FILE "\t" .  join("\t", @totAtoms) . "\t" . $massList{$compound}{$frag};
         if ( $conflictList{$compound}{$frag} ) { print FILE "*"; }
			print FILE "\t" . $firstC . "-" . $lastC . "\n";
      }
   }
   
   print FILE "\n*  Conflicting mass. Cannot uniquely identify carbons in backbone.\n\t- Can correct for NA but not for OBM.\n\t- Should not be used when generating MFA measurement data.\n";
   print FILE "** Side chain binds to a free H, increasing the mass accordingly.\n";
   
   close(FILE)                 or FileSystemError(901, $outFile, $!);
}


#######################
#                     #
#   Public methods.   #
#                     #
#######################

################################################################################
# Subroutine: GenerateAtomList
#
# Generates a list of atom arrays in a hash (indexed by compound key) of hashes
# (indexed by fragment key).
#
################################################################################
sub GenerateAtomList {
   foreach my $compound (sort keys %compounds) {
		my @totAtoms   = @{${$compounds{$compound}}[1]};
		my $multiplier = ${$compounds{$compound}}[2]; 
		
		for(my $i=0; $i<@totAtoms; $i++) {
			$totAtoms[$i] += ($dAgent[$i] * $multiplier);

			# Correct for Hydrogen loss, 1 per dAgent:
			if ( $atomList[$i] eq "H" ) { $totAtoms[$i] -= $multiplier; }
		}
		
		my %list = ();
		foreach my $frag (sort keys %frags) {      
			my @adj = @{$frags{$frag}};

			my @atoms = ();
			for(my $i=0; $i<@totAtoms; $i++) {
            $atoms[$i] = $totAtoms[$i] - $adj[$i];
				
				# For the m302s side chain, add extra binding hydrogen(s):
				if ( $frag eq "m302s") {
					if ( $atomList[$i] eq "H" ) { $atoms[$i] += ${$compounds{$compound}}[3]; }
				}
			}
		 
			$list{$frag} = \@atoms;
		}
		
		$list{'m302'} = \@m302;
		
		$totalAtomList{$compound} = \%list;
	}
}


################################################################################
# Subroutine: GenerateMassList
#
# Generates a list of mass values in a hash (indexed by compound key) of hashes
# (indexed by fragment key).
#
################################################################################
sub GenerateMassList {
   foreach my $compound (sort keys %compounds) {
 		my @frags = sort keys %{$totalAtomList{$compound}};
  
      my %list = ();
      foreach my $frag (@frags) {
			my @totAtoms = @{$totalAtomList{$compound}{$frag}};
				
			my $mass  = 0;
			for(my $i=0; $i<@totAtoms; $i++) { $mass += ($totAtoms[$i] * $atomMass[$i]); }
			
         $list{$frag} = &Trunc($mass);
      }
      
      $massList{$compound} = \%list;
   }
}


################################################################################
# Subroutine: GenerateConflictList
#
# Generates a list of fragments with conflicting mass entries for each compound.
# The list is a hash (indexed by compound key) of hashes (indexed by fragment 
# key).
#
################################################################################
sub GenerateConflictList {
	foreach my $compound (sort keys %compounds) {
		my @frags = keys %{$massList{$compound}};
		
		my @masses = ();
      foreach my $frag (@frags) {
         push(@masses, $massList{$compound}{$frag});
      }
      
      my %list = ();
      foreach my $frag (@frags) {
         my $mass = $massList{$compound}{$frag};
         my $found = 0;
			for (my $i=0; $i<@masses; $i++) {
			   if ( $mass == $masses[$i] ) { $found++; }
			}
         $list{$frag} = $found - 1;
      }
      
      $conflictList{$compound} = \%list;
   }
}


################################################################################
# Subroutine: GetCarbonBackbone
#
# Returns the first and last carbon in the fragments carbon backbone.
#
################################################################################
sub GetCarbonBackbone {
   my ($compound, $fragMass, $lastC) = @_;
   
   my $firstC = 1;

   foreach my $frag (sort keys %frags) {
      # Increment the starting carbon position, assumes keys are in order.	
      if ( $frag eq 'm085' ) { $firstC++; }
      if ( $frag eq 'm302s' ) { $firstC++; }

      if ($fragMass == $massList{$compound}{$frag}) {
         return ($firstC, $lastC);
      }
   }
      
   if ($fragMass == $massList{$compound}{'m302'}) { return (1,2); }
   
   return;
}


################################################################################
# Subroutine: GetTotalAtomList
#
# Return the totalAtomList hash.
#
################################################################################
sub GetTotalAtomList {
   return %totalAtomList;
}


################################################################################
# Subroutine: GetMassList
#
# Return the massList hash.
#
################################################################################
sub GetMassList {
   return %massList;
}


################################################################################
# Subroutine: GetConflictList
#
# Return the conflictList hash.
#
################################################################################
sub GetConflictList {
   return %conflictList;
}


1;